import copy
import json
from typing import Union
from urllib.parse import urlencode

from py_yt.core.constants import requestPayload, searchKey, ResultMode
from py_yt.core.requests import RequestCore
from py_yt.handlers.componenthandler import ComponentHandler


class ChannelSearchCore(RequestCore, ComponentHandler):
    response = None
    responseSource = None
    resultComponents = []

    def __init__(
        self,
        query: str,
        language: str,
        region: str,
        search_preferences: str,
        browse_id: str,
        timeout: int,
    ):
        super().__init__()
        self.query = query
        self.language = language
        self.region = region
        self.browseId = browse_id
        self.searchPreferences = search_preferences
        self.continuationKey = None
        self.timeout = timeout

    async def next(self):
        await self._asyncRequest()
        self._parseChannelSearchSource()
        self.response = self._getChannelSearchComponent(self.response)
        return self.response

    def _parseChannelSearchSource(self) -> None:
        try:
            last_tab = self.response["contents"]["twoColumnBrowseResultsRenderer"][
                "tabs"
            ][-1]
            if "expandableTabRenderer" in last_tab:
                renderer = last_tab["expandableTabRenderer"]
                if "content" in renderer:
                    self.response = renderer["content"]["sectionListRenderer"]["contents"]
                else:
                    self.response = []
            elif "tabRenderer" in last_tab:
                tab_renderer = last_tab["tabRenderer"]
                if "content" in tab_renderer:
                    self.response = tab_renderer["content"]["sectionListRenderer"][
                        "contents"
                    ]
                else:
                    self.response = []
            else:
                self.response = []
        except:
            raise Exception("ERROR: Could not parse YouTube response.")

    def _getRequestBody(self):
        requestBody = copy.deepcopy(requestPayload)
        requestBody["query"] = self.query
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        requestBody["params"] = self.searchPreferences
        requestBody["browseId"] = self.browseId
        self.url = (
            "https://www.youtube.com/youtubei/v1/browse"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            )
        )
        self.data = requestBody

    async def _asyncRequest(self) -> None:
        self._getRequestBody()

        request = await self.asyncPostRequest()
        try:
            self.response = request.json()
        except:
            raise Exception("ERROR: Could not make request.")

    def result(self, mode: int = ResultMode.dict) -> Union[str, dict]:
        """Returns the search result.
        Args:
            mode (int, optional): Sets the type of result. Defaults to ResultMode.dict.
        Returns:
            Union[str, dict]: Returns JSON or dictionary.
        """
        if mode == ResultMode.json:
            return json.dumps({"result": self.response}, indent=4)
        elif mode == ResultMode.dict:
            return {"result": self.response}




from typing import Any, Dict, Optional

from py_yt.core.channelsearch import ChannelSearchCore
from py_yt.core.constants import VideoSortOrder, SearchMode
from py_yt.core.search import SearchCore


class Search(SearchCore):
    """Searches for videos, channels & playlists in YouTube.

    Args:
        query (str): Sets the search query.
        limit (int, optional): Sets limit to the number of results. Defaults to 20.
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = Search('Watermelon Sugar', limit = 1)
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "type": "video",
                    "id": "E07s5ZYygMg",
                    "title": "Harry Styles - Watermelon Sugar (Official Video)",
                    "publishedTime": "6 months ago",
                    "duration": "3:09",
                    "viewCount": {
                        "text": "162,235,006 views",
                        "short": "162M views"
                    },
                    "thumbnails": [
                        {
                            "url": "https://i.ytimg.com/vi/E07s5ZYygMg/hq720.jpg?sqp=-oaymwEjCOgCEMoBSFryq4qpAxUIARUAAAAAGAElAADIQj0AgKJDeAE=&rs=AOn4CLAOWBTE1SDrtrDQ1aWNzpDZ7YiMIw",
                            "width": 360,
                            "height": 202
                        },
                        {
                            "url": "https://i.ytimg.com/vi/E07s5ZYygMg/hq720.jpg?sqp=-oaymwEXCNAFEJQDSFryq4qpAwkIARUAAIhCGAE=&rs=AOn4CLD7U54pGZLPKTuMP-J3kpm4LIDPVg",
                            "width": 720,
                            "height": 404
                        }
                    ],
                    "descriptionSnippet": [
                        {
                            "text": "This video is dedicated to touching. Listen to Harry Styles' new album 'Fine Line' now: https://HStyles.lnk.to/FineLineAY Follow\u00a0..."
                        }
                    ],
                    "channel": {
                        "name": "Harry Styles",
                        "id": "UCZFWPqqPkFlNwIxcpsLOwew",
                        "thumbnails": [
                            {
                                "url": "https://yt3.ggpht.com/a-/AOh14GgNUvHxwlnz4RpHamcGnZF1px13VHj01TPksw=s68-c-k-c0x00ffffff-no-rj-mo",
                                "width": 68,
                                "height": 68
                            }
                        ],
                        "link": "https://www.youtube.com/channel/UCZFWPqqPkFlNwIxcpsLOwew"
                    },
                    "accessibility": {
                        "title": "Harry Styles - Watermelon Sugar (Official Video) by Harry Styles 6 months ago 3 minutes, 9 seconds 162,235,006 views",
                        "duration": "3 minutes, 9 seconds"
                    },
                    "link": "https://www.youtube.com/watch?v=E07s5ZYygMg",
                    "shelfTitle": null
                }
            ]
        }
    """

    def __init__(
        self,
        query: str,
        limit: int = 20,
        language: str = "en",
        region: str = "US",
        timeout: Optional[int] = None,
    ):
        self.searchMode = (True, True, True)
        super().__init__(query, limit, language, region, None, timeout)  # type: ignore

    async def next(self) -> Dict[str, Any]:
        return await self._nextAsync()  # type: ignore


class VideosSearch(SearchCore):
    """Searches for videos in YouTube.

    Args:
        query (str): Sets the search query.
        limit (int, optional): Sets limit to the number of results. Defaults to 20.
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = VideosSearch('Watermelon Sugar', limit = 1)
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "type": "video",
                    "id": "E07s5ZYygMg",
                    "title": "Harry Styles - Watermelon Sugar (Official Video)",
                    "publishedTime": "6 months ago",
                    "duration": "3:09",
                    "viewCount": {
                        "text": "162,235,006 views",
                        "short": "162M views"
                    },
                    "thumbnails": [
                        {
                            "url": "https://i.ytimg.com/vi/E07s5ZYygMg/hq720.jpg?sqp=-oaymwEjCOgCEMoBSFryq4qpAxUIARUAAAAAGAElAADIQj0AgKJDeAE=&rs=AOn4CLAOWBTE1SDrtrDQ1aWNzpDZ7YiMIw",
                            "width": 360,
                            "height": 202
                        },
                        {
                            "url": "https://i.ytimg.com/vi/E07s5ZYygMg/hq720.jpg?sqp=-oaymwEXCNAFEJQDSFryq4qpAwkIARUAAIhCGAE=&rs=AOn4CLD7U54pGZLPKTuMP-J3kpm4LIDPVg",
                            "width": 720,
                            "height": 404
                        }
                    ],
                    "descriptionSnippet": [
                        {
                            "text": "This video is dedicated to touching. Listen to Harry Styles' new album 'Fine Line' now: https://HStyles.lnk.to/FineLineAY Follow\u00a0..."
                        }
                    ],
                    "channel": {
                        "name": "Harry Styles",
                        "id": "UCZFWPqqPkFlNwIxcpsLOwew",
                        "thumbnails": [
                            {
                                "url": "https://yt3.ggpht.com/a-/AOh14GgNUvHxwlnz4RpHamcGnZF1px13VHj01TPksw=s68-c-k-c0x00ffffff-no-rj-mo",
                                "width": 68,
                                "height": 68
                            }
                        ],
                        "link": "https://www.youtube.com/channel/UCZFWPqqPkFlNwIxcpsLOwew"
                    },
                    "accessibility": {
                        "title": "Harry Styles - Watermelon Sugar (Official Video) by Harry Styles 6 months ago 3 minutes, 9 seconds 162,235,006 views",
                        "duration": "3 minutes, 9 seconds"
                    },
                    "link": "https://www.youtube.com/watch?v=E07s5ZYygMg",
                    "shelfTitle": null
                }
            ]
        }
    """

    def __init__(
        self,
        query: str,
        limit: int = 20,
        language: str = "en",
        region: str = "US",
        timeout: Optional[int] = None,
    ):
        self.searchMode = (True, False, False)
        super().__init__(query, limit, language, region, SearchMode.videos, timeout)

    async def next(self) -> Dict[str, Any]:
        return await self._nextAsync()  # type: ignore


class ChannelsSearch(SearchCore):
    """Searches for channels in YouTube.

    Args:
        query (str): Sets the search query.
        limit (int, optional): Sets limit to the number of results. Defaults to 20.
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = ChannelsSearch('Harry Styles', limit = 1)
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "type": "channel",
                    "id": "UCZFWPqqPkFlNwIxcpsLOwew",
                    "title": "Harry Styles",
                    "thumbnails": [
                        {
                            "url": "https://yt3.ggpht.com/ytc/AAUvwnhR81ocC_KalYEk5ItnJcfMBqaiIpuM1B0lJyg4Rw=s88-c-k-c0x00ffffff-no-rj-mo",
                            "width": 88,
                            "height": 88
                        },
                        {
                            "url": "https://yt3.ggpht.com/ytc/AAUvwnhR81ocC_KalYEk5ItnJcfMBqaiIpuM1B0lJyg4Rw=s176-c-k-c0x00ffffff-no-rj-mo",
                            "width": 176,
                            "height": 176
                        }
                    ],
                    "videoCount": "7",
                    "descriptionSnippet": null,
                    "subscribers": "9.25M subscribers",
                    "link": "https://www.youtube.com/channel/UCZFWPqqPkFlNwIxcpsLOwew"
                }
            ]
        }
    """

    def __init__(
        self,
        query: str,
        limit: int = 20,
        language: str = "en",
        region: str = "US",
        timeout: Optional[int] = None,
    ):
        self.searchMode = (False, True, False)
        super().__init__(query, limit, language, region, SearchMode.channels, timeout)  # type: ignore

    async def next(self) -> Dict[str, Any]:
        return await self._nextAsync()  # type: ignore


class PlaylistsSearch(SearchCore):
    """Searches for playlists in YouTube.

    Args:
        query (str): Sets the search query.
        limit (int, optional): Sets limit to the number of results. Defaults to 20.
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = PlaylistsSearch('Harry Styles', limit = 1)
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "type": "playlist",
                    "id": "PL-Rt4gIwHnyvxpEl-9Le0ePztR7WxGDGV",
                    "title": "fine line harry styles full album lyrics",
                    "videoCount": "12",
                    "channel": {
                        "name": "ourmemoriestonight",
                        "id": "UCZCmb5a8LE9LMxW9I3-BFjA",
                        "link": "https://www.youtube.com/channel/UCZCmb5a8LE9LMxW9I3-BFjA"
                    },
                    "thumbnails": [
                        {
                            "url": "https://i.ytimg.com/vi/raTh8Mu5oyM/hqdefault.jpg?sqp=-oaymwEWCKgBEF5IWvKriqkDCQgBFQAAiEIYAQ==&rs=AOn4CLCdCfOQYMrPImHMObdrMcNimKi1PA",
                            "width": 168,
                            "height": 94
                        },
                        {
                            "url": "https://i.ytimg.com/vi/raTh8Mu5oyM/hqdefault.jpg?sqp=-oaymwEWCMQBEG5IWvKriqkDCQgBFQAAiEIYAQ==&rs=AOn4CLDsKmyGH8bkmt9MzZqIoXI4UaduBw",
                            "width": 196,
                            "height": 110
                        },
                        {
                            "url": "https://i.ytimg.com/vi/raTh8Mu5oyM/hqdefault.jpg?sqp=-oaymwEXCPYBEIoBSFryq4qpAwkIARUAAIhCGAE=&rs=AOn4CLD9v7S0KeHLBLr0bF-LrRjYVycUFA",
                            "width": 246,
                            "height": 138
                        },
                        {
                            "url": "https://i.ytimg.com/vi/raTh8Mu5oyM/hqdefault.jpg?sqp=-oaymwEXCNACELwBSFryq4qpAwkIARUAAIhCGAE=&rs=AOn4CLAIzQIVxZsC0PfvLOt-v9UWJ-109Q",
                            "width": 336,
                            "height": 188
                        }
                    ],
                    "link": "https://www.youtube.com/playlist?list=PL-Rt4gIwHnyvxpEl-9Le0ePztR7WxGDGV"
                }
            ]
        }
    """

    def __init__(
        self,
        query: str,
        limit: int = 20,
        language: str = "en",
        region: str = "US",
        timeout: Optional[int] = None,
    ):
        self.searchMode = (False, False, True)
        super().__init__(query, limit, language, region, SearchMode.playlists, timeout)  # type: ignore

    async def next(self) -> Dict[str, Any]:
        return await self._nextAsync()  # type: ignore


class CustomSearch(SearchCore):
    """Performs custom search in YouTube with search filters or sorting orders.
    Few of the predefined filters and sorting orders are:

        1 - SearchMode.videos
        2 - VideoUploadDateFilter.lastHour
        3 - VideoDurationFilter.long
        4 - VideoSortOrder.viewCount

    There are many other to use.
    The value of `sp` parameter in the YouTube search query can be used as a search filter e.g.
    `EgQIBRAB` from https://www.youtube.com/results?search_query=NoCopyrightSounds&sp=EgQIBRAB can be passed as `searchPreferences`, to get videos, which are uploaded this year.

    Args:
        query (str): Sets the search query.
        search_preferences (str): Sets the `sp` query parameter in the YouTube search request.
        limit (int, optional): Sets limit to the number of results. Defaults to 20.
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = CustomSearch('Harry Styles', VideoSortOrder.viewCount, limit = 1)
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "type": "video",
                    "id": "QJO3ROT-A4E",
                    "title": "One Direction - What Makes You Beautiful (Official Video)",
                    "publishedTime": "9 years ago",
                    "duration": "3:27",
                    "viewCount": {
                        "text": "1,212,146,802 views",
                        "short": "1.2B views"
                    },
                    "thumbnails": [
                        {
                            "url": "https://i.ytimg.com/vi/QJO3ROT-A4E/hq720.jpg?sqp=-oaymwEjCOgCEMoBSFryq4qpAxUIARUAAAAAGAElAADIQj0AgKJDeAE=&rs=AOn4CLDeFKrH99gmpnvKyG4czdd__YRDkw",
                            "width": 360,
                            "height": 202
                        },
                        {
                            "url": "https://i.ytimg.com/vi/QJO3ROT-A4E/hq720.jpg?sqp=-oaymwEXCNAFEJQDSFryq4qpAwkIARUAAIhCGAE=&rs=AOn4CLBJ_wUjsRFXGsbvRpwYpSLlsGmbkw",
                            "width": 720,
                            "height": 404
                        }
                    ],
                    "descriptionSnippet": [
                        {
                            "text": "One Direction \u2013 What Makes You Beautiful (Official Video) Follow on Spotify - https://1D.lnk.to/Spotify Listen on Apple Music\u00a0..."
                        }
                    ],
                    "channel": {
                        "name": "One Direction",
                        "id": "UCb2HGwORFBo94DmRx4oLzow",
                        "thumbnails": [
                            {
                                "url": "https://yt3.ggpht.com/a-/AOh14Gj3SMvtIAvVNUrHWFTJFubPN7qozzPl5gFkoA=s68-c-k-c0x00ffffff-no-rj-mo",
                                "width": 68,
                                "height": 68
                            }
                        ],
                        "link": "https://www.youtube.com/channel/UCb2HGwORFBo94DmRx4oLzow"
                    },
                    "accessibility": {
                        "title": "One Direction - What Makes You Beautiful (Official Video) by One Direction 9 years ago 3 minutes, 27 seconds 1,212,146,802 views",
                        "duration": "3 minutes, 27 seconds"
                    },
                    "link": "https://www.youtube.com/watch?v=QJO3ROT-A4E",
                    "shelfTitle": null
                }
            ]
        }
    """

    def __init__(
        self,
        query: str,
        search_preferences: str,
        limit: int = 20,
        language: str = "en",
        region: str = "US",
        timeout: Optional[int] = None,
    ):
        self.searchMode = (True, True, True)
        super().__init__(query, limit, language, region, search_preferences, timeout)

    async def next(self) -> Dict[str, Any]:
        return await self._nextAsync()


class ChannelSearch(ChannelSearchCore):
    """Searches for videos in specific channel in YouTube.

    Args:
        query (str): Sets the search query.
        browse_id (str): Channel ID
        language (str, optional): Sets the result language. Defaults to 'en'.
        region (str, optional): Sets the result region. Defaults to 'US'.

    Examples:
        Calling `result` method gives the search result.

        >>> search = ChannelSearch('Watermelon Sugar', "UCZFWPqqPkFlNwIxcpsLOwew")
        >>> result = await search.next()
        >>> print(result)
        {
            "result": [
                {
                    "id": "WMcIfZuRuU8",
                    "thumbnails": {
                        "normal": [
                            {
                                "url": "https://i.ytimg.com/vi/WMcIfZuRuU8/hqdefault.jpg?sqp=-oaymwEbCKgBEF5IVfKriqkDDggBFQAAiEIYAXABwAEG&rs=AOn4CLClFg6C1r5NfTQy7TYUq6X5qHUmPA",
                                "width": 168,
                                "height": 94
                            },
                            {
                                "url": "https://i.ytimg.com/vi/WMcIfZuRuU8/hqdefault.jpg?sqp=-oaymwEbCMQBEG5IVfKriqkDDggBFQAAiEIYAXABwAEG&rs=AOn4CLAoOyftwY0jLV4geWb5hejULYp3Zw",
                                "width": 196,
                                "height": 110
                            },
                            {
                                "url": "https://i.ytimg.com/vi/WMcIfZuRuU8/hqdefault.jpg?sqp=-oaymwEcCPYBEIoBSFXyq4qpAw4IARUAAIhCGAFwAcABBg==&rs=AOn4CLCdqkhn7JDwLvRtTNx3jq-olz7k-Q",
                                "width": 246,
                                "height": 138
                            },
                            {
                                "url": "https://i.ytimg.com/vi/WMcIfZuRuU8/hqdefault.jpg?sqp=-oaymwEcCNACELwBSFXyq4qpAw4IARUAAIhCGAFwAcABBg==&rs=AOn4CLAhYedsqBFKI0Ra2qzIv9cVoZhfKQ",
                                "width": 336,
                                "height": 188
                            }
                        ],
                        "rich": null
                    },
                    "title": "Harry Styles \u2013 Watermelon Sugar (Lost Tour Visual)",
                    "descriptionSnippet": "This video is dedicated to touching.\nListen to Harry Styles\u2019 new album \u2018Fine Line\u2019 now: https://HStyles.lnk.to/FineLineAY \n\nFollow Harry Styles:\nFacebook: https://HarryStyles.lnk.to/followFI...",
                    "uri": "/watch?v=WMcIfZuRuU8",
                    "views": {
                        "precise": "3,888,287 views",
                        "simple": "3.8M views",
                        "approximate": "3.8 million views"
                    },
                    "duration": {
                        "simpleText": "2:55",
                        "text": "2 minutes, 55 seconds"
                    },
                    "published": "10 months ago",
                    "channel": {
                        "name": "Harry Styles",
                        "thumbnails": [
                            {
                                "url": "https://yt3.ggpht.com/ytc/AAUvwnhR81ocC_KalYEk5ItnJcfMBqaiIpuM1B0lJyg4Rw=s88-c-k-c0x00ffffff-no-rj",
                                "width": 68,
                                "height": 68
                            }
                        ]
                    },
                    "type": "video"
                },
            ]
        }
    """

    def __init__(
        self,
        query: str,
        browse_id: str,
        language: str = "en",
        region: str = "US",
        search_preferences: str = "EgZzZWFyY2g%3D",
        timeout: Optional[int] = None,
    ):
        super().__init__(query, language, region, search_preferences, browse_id, timeout)  # type: ignore



import copy
import json
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from py_yt.core.constants import (
    requestPayload,
    searchKey,
    userAgent,
    contentPath,
    itemSectionKey,
    continuationItemKey,
    continuationKeyPath,
    fallbackContentPath,
    continuationContentPath,
)
from py_yt.handlers.componenthandler import ComponentHandler


class RequestHandler(ComponentHandler):
    def _makeRequest(self) -> None:
        requestBody = copy.deepcopy(requestPayload)
        requestBody["query"] = self.query
        requestBody["client"] = {
            "hl": self.language,
            "gl": self.region,
        }
        if self.searchPreferences:
            requestBody["params"] = self.searchPreferences
        if self.continuationKey:
            requestBody["continuation"] = self.continuationKey
        requestBodyBytes = json.dumps(requestBody).encode("utf_8")
        request = Request(
            "https://www.youtube.com/youtubei/v1/search"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            ),
            data=requestBodyBytes,
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Content-Length": len(requestBodyBytes),
                "User-Agent": userAgent,
            },
        )
        try:
            self.response = (
                urlopen(request, timeout=self.timeout).read().decode("utf_8")
            )
        except:
            raise Exception("ERROR: Could not make request.")

    def _parseSource(self) -> None:
        try:
            if not self.continuationKey:
                responseContent = self._getValue(json.loads(self.response), contentPath)
            else:
                responseContent = self._getValue(
                    json.loads(self.response), continuationContentPath
                )
            if responseContent:
                for element in responseContent:
                    if itemSectionKey in element.keys():
                        self.responseSource = self._getValue(
                            element, [itemSectionKey, "contents"]
                        )
                    if continuationItemKey in element.keys():
                        self.continuationKey = self._getValue(
                            element, continuationKeyPath
                        )
            else:
                self.responseSource = self._getValue(
                    json.loads(self.response), fallbackContentPath
                )
                self.continuationKey = self._getValue(
                    self.responseSource[-1], continuationKeyPath
                )
        except:
            raise Exception("ERROR: Could not parse YouTube response.")



requestPayload = {
    "context": {
        "client": {
            "clientName": "WEB",
            "clientVersion": "2.20210224.06.00",
            "newVisitorCookie": True,
        },
        "user": {
            "lockedSafetyMode": False,
        },
    }
}

userAgent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36"

videoElementKey = "videoRenderer"
channelElementKey = "channelRenderer"
playlistElementKey = "playlistRenderer"
shelfElementKey = "shelfRenderer"
itemSectionKey = "itemSectionRenderer"
continuationItemKey = "continuationItemRenderer"
playerResponseKey = "playerResponse"
richItemKey = "richItemRenderer"
hashtagElementKey = "hashtagTileRenderer"
hashtagBrowseKey = "FEhashtag"
hashtagVideosPath = [
    "contents",
    "twoColumnBrowseResultsRenderer",
    "tabs",
    0,
    "tabRenderer",
    "content",
    "richGridRenderer",
    "contents",
]
hashtagContinuationVideosPath = [
    "onResponseReceivedActions",
    0,
    "appendContinuationItemsAction",
    "continuationItems",
]
searchKey = "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8"
contentPath = [
    "contents",
    "twoColumnSearchResultsRenderer",
    "primaryContents",
    "sectionListRenderer",
    "contents",
]
fallbackContentPath = [
    "contents",
    "twoColumnSearchResultsRenderer",
    "primaryContents",
    "richGridRenderer",
    "contents",
]
continuationContentPath = [
    "onResponseReceivedCommands",
    0,
    "appendContinuationItemsAction",
    "continuationItems",
]
continuationKeyPath = [
    "continuationItemRenderer",
    "continuationEndpoint",
    "continuationCommand",
    "token",
]
playlistInfoPath = ["response", "sidebar", "playlistSidebarRenderer", "items"]
playlistVideosPath = [
    "response",
    "contents",
    "twoColumnBrowseResultsRenderer",
    "tabs",
    0,
    "tabRenderer",
    "content",
    "sectionListRenderer",
    "contents",
    0,
    "itemSectionRenderer",
    "contents",
    0,
    "playlistVideoListRenderer",
    "contents",
]
playlistPrimaryInfoKey = "playlistSidebarPrimaryInfoRenderer"
playlistSecondaryInfoKey = "playlistSidebarSecondaryInfoRenderer"
playlistVideoKey = "playlistVideoRenderer"


class ResultMode:
    json = 0
    dict = 1


class SearchMode:
    videos = "EgIQAQ%3D%3D"
    channels = "EgIQAg%3D%3D"
    playlists = "EgIQAw%3D%3D"
    livestreams = "EgJAAQ%3D%3D"


class VideoUploadDateFilter:
    lastHour = "EgQIARAB"
    today = "EgQIAhAB"
    thisWeek = "EgQIAxAB"
    thisMonth = "EgQIBBAB"
    thisYear = "EgQIBRAB"


class VideoDurationFilter:
    short = "EgQQARgB"
    long = "EgQQARgC"


class VideoSortOrder:
    relevance = "CAASAhAB"
    uploadDate = "CAISAhAB"
    viewCount = "CAMSAhAB"
    rating = "CAESAhAB"


class ChannelRequestType:
    info = "EgVhYm91dA%3D%3D"
    playlists = "EglwbGF5bGlzdHMYAyABcAA%3D"



import copy
from urllib.parse import urlencode

from py_yt.core.componenthandler import getVideoId, getValue
from py_yt.core.constants import searchKey, requestPayload
from py_yt.core.requests import RequestCore


class TranscriptCore(RequestCore):
    def __init__(self, videoLink: str, key: str):
        super().__init__()
        self.videoLink = videoLink
        self.key = key

    def prepare_params_request(self):
        self.url = (
            "https://www.youtube.com/youtubei/v1/next"
            + "?"
            + urlencode({"key": searchKey, "prettyPrint": "false"})
        )
        self.data = copy.deepcopy(requestPayload)
        self.data["videoId"] = getVideoId(self.videoLink)

    def extract_continuation_key(self, r):
        j = r.json()
        panels = getValue(j, ["engagementPanels"])
        if not panels:
            raise Exception(
                "Failed to create first request - No engagementPanels is present."
            )
        key = ""
        for panel in panels:
            panel = panel["engagementPanelSectionListRenderer"]
            if (
                getValue(panel, ["targetId"])
                == "engagement-panel-searchable-transcript"
            ):
                key = getValue(
                    panel,
                    [
                        "content",
                        "continuationItemRenderer",
                        "continuationEndpoint",
                        "getTranscriptEndpoint",
                        "params",
                    ],
                )
        if key == "" or not key:
            self.result = {"segments": [], "languages": []}
            return True
        self.key = key
        return False

    def prepare_transcript_request(self):
        self.url = (
            "https://www.youtube.com/youtubei/v1/get_transcript"
            + "?"
            + urlencode({"key": searchKey, "prettyPrint": "false"})
        )
        # clientVersion must be newer than in requestPayload
        self.data = {
            "context": {
                "client": {
                    "clientName": "WEB",
                    "clientVersion": "2.20220318.00.00",
                    "newVisitorCookie": True,
                },
                "user": {
                    "lockedSafetyMode": False,
                },
            },
            "params": self.key,
        }

    def extract_transcript(self):
        response = self.data
        transcripts = getValue(
            response,
            [
                "actions",
                0,
                "updateEngagementPanelAction",
                "content",
                "transcriptRenderer",
                "content",
                "transcriptSearchPanelRenderer",
                "body",
                "transcriptSegmentListRenderer",
                "initialSegments",
            ],
        )
        segments = []
        languages = []
        if transcripts:
            for segment in transcripts:
                segment = getValue(segment, ["transcriptSegmentRenderer"])
                if segment:
                    j = {
                        "startMs": getValue(segment, ["startMs"]),
                        "endMs": getValue(segment, ["endMs"]),
                        "text": getValue(segment, ["snippet", "runs", 0, "text"]),
                        "startTime": getValue(
                            segment, ["startTimeText", "simpleText"]
                        ),
                    }
                    segments.append(j)
        langs = getValue(
            response,
            [
                "actions",
                0,
                "updateEngagementPanelAction",
                "content",
                "transcriptRenderer",
                "content",
                "transcriptSearchPanelRenderer",
                "footer",
                "transcriptFooterRenderer",
                "languageMenu",
                "sortFilterSubMenuRenderer",
                "subMenuItems",
            ],
        )
        if langs:
            for language in langs:
                j = {
                    "params": getValue(
                        language,
                        ["continuation", "reloadContinuationData", "continuation"],
                    ),
                    "selected": getValue(language, ["selected"]),
                    "title": getValue(language, ["title"]),
                }
                languages.append(j)
        self.result = {"segments": segments, "languages": languages}

    async def async_create(self):
        if not self.key:
            self.prepare_params_request()
            r = await self.asyncPostRequest()
            end = self.extract_continuation_key(r)
            if end:
                return
        self.prepare_transcript_request()
        response = await self.asyncPostRequest()
        if response:
            self.data = response.json()
            self.extract_transcript()



import copy
import json
import re
from typing import Iterable, Mapping, TypeVar, Union, List
from urllib.parse import urlencode

from py_yt.core.constants import (
    ResultMode,
    searchKey,
    requestPayload,
    playlistVideoKey,
    continuationKeyPath,
    playlistPrimaryInfoKey,
    playlistSecondaryInfoKey,
    continuationItemKey,
)
from py_yt.core.requests import RequestCore

K = TypeVar("K")
T = TypeVar("T")


class PlaylistCore(RequestCore):
    playlistComponent = None
    result = None
    continuationKey = None

    def __init__(
        self, playlist_link: str, componentMode: str, result_mode: int, timeout: int
    ):
        super().__init__()
        self.componentMode = componentMode
        self.resultMode = result_mode
        self.timeout = timeout
        self.url = playlist_link

    def post_processing(self):
        self.__parseSource()
        self.__getComponents()
        if self.resultMode == ResultMode.json:
            self.result = json.dumps(self.playlistComponent, indent=4)
        else:
            self.result = self.playlistComponent

    async def async_create(self):
        statusCode = await self.__makeAsyncRequest()
        if statusCode == 200:
            self.post_processing()
        else:
            raise Exception("ERROR: Invalid status code.")

    def next_post_processing(self):
        self.__parseSource()
        self.__getNextComponents()
        if self.resultMode == ResultMode.json:
            self.result = json.dumps(self.playlistComponent, indent=4)
        else:
            self.result = self.playlistComponent

    async def _async_next(self):
        if self.continuationKey:
            self.prepare_next_request()
            statusCode = await self.asyncPostRequest()
            self.response = statusCode.text
            if statusCode.status_code == 200:
                self.next_post_processing()
            else:
                raise Exception("ERROR: Invalid status code.")
        else:
            await self.async_create()

    def prepare_first_request(self):
        self.url.strip("/")

        _id = re.search(r"(?<=list=)([a-zA-Z0-9+/=_-]+)", self.url).group()
        browseId = "VL" + _id if not _id.startswith("VL") else _id

        self.url = (
            "https://www.youtube.com/youtubei/v1/browse"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            )
        )
        self.data = {
            "browseId": browseId,
        }
        self.data.update(copy.deepcopy(requestPayload))

    async def __makeAsyncRequest(self) -> int:
        self.prepare_first_request()
        request = await self.asyncPostRequest()
        self.response = request.text
        return request.status_code

    def prepare_next_request(self):
        requestBody = copy.deepcopy(requestPayload)
        requestBody["continuation"] = self.continuationKey
        self.data = requestBody
        self.url = (
            "https://www.youtube.com/youtubei/v1/browse"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                }
            )
        )

    def __parseSource(self) -> None:
        try:
            self.responseSource = json.loads(self.response)
        except:
            raise Exception("ERROR: Could not parse YouTube response.")

    def __getComponents(self) -> None:
        # print(self.responseSource)
        sidebar = self.responseSource["sidebar"]["playlistSidebarRenderer"]["items"]
        inforenderer = sidebar[0]["playlistSidebarPrimaryInfoRenderer"]
        channel_details_available = len(sidebar) != 1
        channelrenderer = (
            sidebar[1]["playlistSidebarSecondaryInfoRenderer"]["videoOwner"][
                "videoOwnerRenderer"
            ]
            if channel_details_available
            else None
        )
        videorenderer: list = self.__getFirstValue(
            self.responseSource,
            [
                "contents",
                "twoColumnBrowseResultsRenderer",
                "tabs",
                None,
                "tabRenderer",
                "content",
                "sectionListRenderer",
                "contents",
                None,
                "itemSectionRenderer",
                "contents",
                None,
                "playlistVideoListRenderer",
                "contents",
            ],
        )
        videos = []
        for video in videorenderer:
            try:
                video = video["playlistVideoRenderer"]
                j = {
                    "id": self.__getValue(video, ["videoId"]),
                    "thumbnails": self.__getValue(video, ["thumbnail", "thumbnails"]),
                    "title": self.__getValue(video, ["title", "runs", 0, "text"]),
                    "channel": {
                        "name": self.__getValue(
                            video, ["shortBylineText", "runs", 0, "text"]
                        ),
                        "id": self.__getValue(
                            video,
                            [
                                "shortBylineText",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "browseId",
                            ],
                        ),
                        "link": self.__getValue(
                            video,
                            [
                                "shortBylineText",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "canonicalBaseUrl",
                            ],
                        ),
                    },
                    "duration": self.__getValue(video, ["lengthText", "simpleText"]),
                    "accessibility": {
                        "title": self.__getValue(
                            video,
                            ["title", "accessibility", "accessibilityData", "label"],
                        ),
                        "duration": self.__getValue(
                            video,
                            [
                                "lengthText",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                    },
                    "link": "https://www.youtube.com"
                    + self.__getValue(
                        video,
                        [
                            "navigationEndpoint",
                            "commandMetadata",
                            "webCommandMetadata",
                            "url",
                        ],
                    ),
                    "isPlayable": self.__getValue(video, ["isPlayable"]),
                }
                videos.append(j)
            except:
                pass

        playlistElement = {
            "info": {
                "id": self.__getValue(
                    inforenderer,
                    [
                        "title",
                        "runs",
                        0,
                        "navigationEndpoint",
                        "watchEndpoint",
                        "playlistId",
                    ],
                ),
                "thumbnails": self.__getValue(
                    inforenderer,
                    [
                        "thumbnailRenderer",
                        "playlistVideoThumbnailRenderer",
                        "thumbnail",
                        "thumbnails",
                    ],
                ),
                "title": self.__getValue(inforenderer, ["title", "runs", 0, "text"]),
                "videoCount": self.__getValue(
                    inforenderer, ["stats", 0, "runs", 0, "text"]
                ),
                "viewCount": self.__getValue(inforenderer, ["stats", 1, "simpleText"]),
                "link": self.__getValue(
                    self.responseSource,
                    ["microformat", "microformatDataRenderer", "urlCanonical"],
                ),
                "channel": {
                    "id": (
                        self.__getValue(
                            channelrenderer,
                            [
                                "title",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "browseId",
                            ],
                        )
                        if channel_details_available
                        else None
                    ),
                    "name": (
                        self.__getValue(channelrenderer, ["title", "runs", 0, "text"])
                        if channel_details_available
                        else None
                    ),
                    "detailsAvailable": channel_details_available,
                    "link": (
                        "https://www.youtube.com"
                        + self.__getValue(
                            channelrenderer,
                            [
                                "title",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "canonicalBaseUrl",
                            ],
                        )
                        if channel_details_available
                        else None
                    ),
                    "thumbnails": (
                        self.__getValue(channelrenderer, ["thumbnail", "thumbnails"])
                        if channel_details_available
                        else None
                    ),
                },
            },
            "videos": videos,
        }
        if self.componentMode == "getInfo":
            self.playlistComponent = playlistElement["info"]
        elif self.componentMode == "getVideos":
            self.playlistComponent = {"videos": videos}
        else:
            self.playlistComponent = playlistElement
        self.continuationKey = self.__getValue(
            videorenderer,
            [
                -1,
                "continuationItemRenderer",
                "continuationEndpoint",
                "continuationCommand",
                "token",
            ],
        )

    def __getNextComponents(self) -> None:
        self.continuationKey = None
        playlistComponent = {
            "videos": [],
        }
        continuationElements = self.__getValue(
            self.responseSource,
            [
                "onResponseReceivedActions",
                0,
                "appendContinuationItemsAction",
                "continuationItems",
            ],
        )
        if continuationElements is None:
            # YouTube Backend issue - See https://github.com/alexmercerind/youtube-search-python/issues/157
            return
        for videoElement in continuationElements:
            if playlistVideoKey in videoElement.keys():
                videoComponent = {
                    "id": self.__getValue(videoElement, [playlistVideoKey, "videoId"]),
                    "title": self.__getValue(
                        videoElement, [playlistVideoKey, "title", "runs", 0, "text"]
                    ),
                    "thumbnails": self.__getValue(
                        videoElement, [playlistVideoKey, "thumbnail", "thumbnails"]
                    ),
                    "link": "https://www.youtube.com"
                    + self.__getValue(
                        videoElement,
                        [
                            playlistVideoKey,
                            "navigationEndpoint",
                            "commandMetadata",
                            "webCommandMetadata",
                            "url",
                        ],
                    ),
                    "channel": {
                        "name": self.__getValue(
                            videoElement,
                            [playlistVideoKey, "shortBylineText", "runs", 0, "text"],
                        ),
                        "id": self.__getValue(
                            videoElement,
                            [
                                playlistVideoKey,
                                "shortBylineText",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "browseId",
                            ],
                        ),
                        "link": "https://www.youtube.com"
                        + self.__getValue(
                            videoElement,
                            [
                                playlistVideoKey,
                                "shortBylineText",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "browseEndpoint",
                                "canonicalBaseUrl",
                            ],
                        ),
                    },
                    "duration": self.__getValue(
                        videoElement, [playlistVideoKey, "lengthText", "simpleText"]
                    ),
                    "accessibility": {
                        "title": self.__getValue(
                            videoElement,
                            [
                                playlistVideoKey,
                                "title",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                        "duration": self.__getValue(
                            videoElement,
                            [
                                playlistVideoKey,
                                "lengthText",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                    },
                }
                playlistComponent["videos"].append(videoComponent)
            self.continuationKey = self.__getValue(videoElement, continuationKeyPath)
        self.playlistComponent["videos"].extend(playlistComponent["videos"])

    def __getPlaylistComponent(self, element: dict, mode: str) -> dict:
        playlistComponent = {}
        if mode in ["getInfo", None]:
            for infoElement in element["info"]:
                if playlistPrimaryInfoKey in infoElement.keys():
                    component = {
                        "id": self.__getValue(
                            infoElement,
                            [
                                playlistPrimaryInfoKey,
                                "title",
                                "runs",
                                0,
                                "navigationEndpoint",
                                "watchEndpoint",
                                "playlistId",
                            ],
                        ),
                        "title": self.__getValue(
                            infoElement,
                            [playlistPrimaryInfoKey, "title", "runs", 0, "text"],
                        ),
                        "videoCount": self.__getValue(
                            infoElement,
                            [playlistPrimaryInfoKey, "stats", 0, "runs", 0, "text"],
                        ),
                        "viewCount": self.__getValue(
                            infoElement,
                            [playlistPrimaryInfoKey, "stats", 1, "simpleText"],
                        ),
                        "thumbnails": self.__getValue(
                            infoElement,
                            [
                                playlistPrimaryInfoKey,
                                "thumbnailRenderer",
                                "playlistVideoThumbnailRenderer",
                                "thumbnail",
                            ],
                        ),
                    }
                    if not component["thumbnails"]:
                        component["thumbnails"] = (
                            self.__getValue(
                                infoElement,
                                [
                                    playlistPrimaryInfoKey,
                                    "thumbnailRenderer",
                                    "playlistCustomThumbnailRenderer",
                                    "thumbnail",
                                    "thumbnails",
                                ],
                            ),
                        )
                    component["link"] = (
                        "https://www.youtube.com/playlist?list=" + component["id"]
                    )
                    playlistComponent.update(component)
                if playlistSecondaryInfoKey in infoElement.keys():
                    component = {
                        "channel": {
                            "name": self.__getValue(
                                infoElement,
                                [
                                    playlistSecondaryInfoKey,
                                    "videoOwner",
                                    "videoOwnerRenderer",
                                    "title",
                                    "runs",
                                    0,
                                    "text",
                                ],
                            ),
                            "id": self.__getValue(
                                infoElement,
                                [
                                    playlistSecondaryInfoKey,
                                    "videoOwner",
                                    "videoOwnerRenderer",
                                    "title",
                                    "runs",
                                    0,
                                    "navigationEndpoint",
                                    "browseEndpoint",
                                    "browseId",
                                ],
                            ),
                            "thumbnails": self.__getValue(
                                infoElement,
                                [
                                    playlistSecondaryInfoKey,
                                    "videoOwner",
                                    "videoOwnerRenderer",
                                    "thumbnail",
                                    "thumbnails",
                                ],
                            ),
                        },
                    }
                    component["channel"]["link"] = (
                        "https://www.youtube.com/channel/" + component["channel"]["id"]
                    )
                    playlistComponent.update(component)
        if mode in ["getVideos", None]:
            self.continuationKey = None
            playlistComponent["videos"] = []
            for videoElement in element["videos"]:
                if playlistVideoKey in videoElement.keys():
                    videoComponent = {
                        "id": self.__getValue(
                            videoElement, [playlistVideoKey, "videoId"]
                        ),
                        "title": self.__getValue(
                            videoElement, [playlistVideoKey, "title", "runs", 0, "text"]
                        ),
                        "thumbnails": self.__getValue(
                            videoElement, [playlistVideoKey, "thumbnail", "thumbnails"]
                        ),
                        "channel": {
                            "name": self.__getValue(
                                videoElement,
                                [
                                    playlistVideoKey,
                                    "shortBylineText",
                                    "runs",
                                    0,
                                    "text",
                                ],
                            ),
                            "id": self.__getValue(
                                videoElement,
                                [
                                    playlistVideoKey,
                                    "shortBylineText",
                                    "runs",
                                    0,
                                    "navigationEndpoint",
                                    "browseEndpoint",
                                    "browseId",
                                ],
                            ),
                        },
                        "duration": self.__getValue(
                            videoElement, [playlistVideoKey, "lengthText", "simpleText"]
                        ),
                        "accessibility": {
                            "title": self.__getValue(
                                videoElement,
                                [
                                    playlistVideoKey,
                                    "title",
                                    "accessibility",
                                    "accessibilityData",
                                    "label",
                                ],
                            ),
                            "duration": self.__getValue(
                                videoElement,
                                [
                                    playlistVideoKey,
                                    "lengthText",
                                    "accessibility",
                                    "accessibilityData",
                                    "label",
                                ],
                            ),
                        },
                    }
                    videoComponent["link"] = (
                        "https://www.youtube.com/watch?v=" + videoComponent["id"]
                    )
                    videoComponent["channel"]["link"] = (
                        "https://www.youtube.com/channel/"
                        + videoComponent["channel"]["id"]
                    )
                    playlistComponent["videos"].append(videoComponent)
                if continuationItemKey in videoElement.keys():
                    self.continuationKey = self.__getValue(
                        videoElement, continuationKeyPath
                    )
        return playlistComponent

    def __result(self, mode: int) -> Union[dict, str]:
        if mode == ResultMode.dict:
            return self.playlistComponent
        elif mode == ResultMode.json:
            return json.dumps(self.playlistComponent, indent=4)

    def __getValue(
        self, source: dict, path: Iterable[str]
    ) -> Union[str, int, dict, None]:
        value = source
        for key in path:
            if type(key) is str:
                if key in value.keys():
                    value = value[key]
                else:
                    value = None
                    break
            elif type(key) is int:
                if len(value) != 0:
                    value = value[key]
                else:
                    value = None
                    break
        return value

    def __getAllWithKey(self, source: Iterable[Mapping[K, T]], key: K) -> Iterable[T]:
        for item in source:
            if key in item:
                yield item[key]

    def __getValueEx(
        self, source: dict, path: List[str]
    ) -> Iterable[Union[str, int, dict, None]]:
        if len(path) <= 0:
            yield source
            return
        key = path[0]
        upcoming = path[1:]
        if key is None:
            following_key = upcoming[0]
            upcoming = upcoming[1:]
            if following_key is None:
                raise Exception(
                    "Cannot search for a key twice consecutive or at the end with no key given"
                )
            values = self.__getAllWithKey(source, following_key)
            for val in values:
                yield from self.__getValueEx(val, path=upcoming)
        else:
            val = self.__getValue(source, path=[key])
            yield from self.__getValueEx(val, path=upcoming)

    def __getFirstValue(
        self, source: dict, path: Iterable[str]
    ) -> Union[str, int, dict, list, None]:
        values = self.__getValueEx(source, list(path))
        for val in values:
            if val is not None:
                return val
        return None



import copy
from urllib.parse import urlencode

from py_yt.core.componenthandler import getValue
from py_yt.core.constants import searchKey, requestPayload
from py_yt.core.requests import RequestCore


class ChannelCore(RequestCore):
    def __init__(self, channel_id: str, request_params: str):
        super().__init__()
        self.browseId = channel_id
        self.params = request_params
        self.result = {}
        self.continuation = None

    def prepare_request(self):
        self.url = (
            "https://www.youtube.com/youtubei/v1/browse"
            + "?"
            + urlencode({"key": searchKey, "prettyPrint": "false"})
        )
        self.data = copy.deepcopy(requestPayload)
        if not self.continuation:
            self.data["params"] = self.params
            self.data["browseId"] = self.browseId
        else:
            self.data["continuation"] = self.continuation

    def playlist_parse(self, i) -> dict:
        return {
            "id": getValue(i, ["playlistId"]),
            "thumbnails": getValue(i, ["thumbnail", "thumbnails"]),
            "title": getValue(i, ["title", "runs", 0, "text"]),
            "videoCount": getValue(i, ["videoCountShortText", "simpleText"]),
            "lastEdited": getValue(i, ["publishedTimeText", "simpleText"]),
        }

    def parse_response(self):
        response = self.data.json()

        thumbnails = []
        try:
            thumbnails.extend(
                getValue(
                    response,
                    ["header", "c4TabbedHeaderRenderer", "avatar", "thumbnails"],
                )
            )
        except:
            pass
        try:
            thumbnails.extend(
                getValue(
                    response,
                    ["metadata", "channelMetadataRenderer", "avatar", "thumbnails"],
                )
            )
        except:
            pass
        try:
            thumbnails.extend(
                getValue(
                    response,
                    [
                        "microformat",
                        "microformatDataRenderer",
                        "thumbnail",
                        "thumbnails",
                    ],
                )
            )
        except:
            pass

        tabData: dict = {}
        playlists: list = []

        for tab in getValue(
            response, ["contents", "twoColumnBrowseResultsRenderer", "tabs"]
        ):
            tab: dict
            title = getValue(tab, ["tabRenderer", "title"])
            if title == "Playlists":
                playlist = getValue(
                    tab,
                    [
                        "tabRenderer",
                        "content",
                        "sectionListRenderer",
                        "contents",
                        0,
                        "itemSectionRenderer",
                        "contents",
                        0,
                        "gridRenderer",
                        "items",
                    ],
                )
                if playlist is not None and getValue(
                    playlist, [0, "gridPlaylistRenderer"]
                ):
                    for i in playlist:
                        if getValue(i, ["continuationItemRenderer"]):
                            self.continuation = getValue(
                                i,
                                [
                                    "continuationItemRenderer",
                                    "continuationEndpoint",
                                    "continuationCommand",
                                    "token",
                                ],
                            )
                            break
                        i: dict = i["gridPlaylistRenderer"]
                        playlists.append(self.playlist_parse(i))
            elif title == "About":
                tabData = tab["tabRenderer"]

        metadata = getValue(
            tabData,
            [
                "content",
                "sectionListRenderer",
                "contents",
                0,
                "itemSectionRenderer",
                "contents",
                0,
                "channelAboutFullMetadataRenderer",
            ],
        )

        self.result = {
            "id": getValue(
                response, ["metadata", "channelMetadataRenderer", "externalId"]
            ),
            "url": getValue(
                response, ["metadata", "channelMetadataRenderer", "channelUrl"]
            ),
            "description": getValue(
                response, ["metadata", "channelMetadataRenderer", "description"]
            ),
            "title": getValue(
                response, ["metadata", "channelMetadataRenderer", "title"]
            ),
            "banners": getValue(
                response, ["header", "c4TabbedHeaderRenderer", "banner", "thumbnails"]
            ),
            "subscribers": {
                "simpleText": getValue(
                    response,
                    [
                        "header",
                        "c4TabbedHeaderRenderer",
                        "subscriberCountText",
                        "simpleText",
                    ],
                ),
                "label": getValue(
                    response,
                    [
                        "header",
                        "c4TabbedHeaderRenderer",
                        "subscriberCountText",
                        "accessibility",
                        "accessibilityData",
                        "label",
                    ],
                ),
            },
            "thumbnails": thumbnails,
            "availableCountryCodes": getValue(
                response,
                ["metadata", "channelMetadataRenderer", "availableCountryCodes"],
            ),
            "isFamilySafe": getValue(
                response, ["metadata", "channelMetadataRenderer", "isFamilySafe"]
            ),
            "keywords": getValue(
                response, ["metadata", "channelMetadataRenderer", "keywords"]
            ),
            "tags": getValue(
                response, ["microformat", "microformatDataRenderer", "tags"]
            ),
            "views": (
                getValue(metadata, ["viewCountText", "simpleText"])
                if metadata
                else None
            ),
            "joinedDate": (
                getValue(metadata, ["joinedDateText", "runs", -1, "text"])
                if metadata
                else None
            ),
            "country": (
                getValue(metadata, ["country", "simpleText"]) if metadata else None
            ),
            "playlists": playlists,
        }

    def parse_next_response(self):
        response = self.data.json()

        self.continuation = None

        response = getValue(
            response,
            [
                "onResponseReceivedActions",
                0,
                "appendContinuationItemsAction",
                "continuationItems",
            ],
        )
        for i in response:
            if getValue(i, ["continuationItemRenderer"]):
                self.continuation = getValue(
                    i,
                    [
                        "continuationItemRenderer",
                        "continuationEndpoint",
                        "continuationCommand",
                        "token",
                    ],
                )
                break
            elif getValue(i, ["gridPlaylistRenderer"]):
                self.result["playlists"].append(
                    self.playlist_parse(getValue(i, ["gridPlaylistRenderer"]))
                )
            # TODO: Handle other types like gridShowRenderer

    async def async_next(self):
        if not self.continuation:
            return
        self.prepare_request()
        self.data = await self.asyncPostRequest()
        self.parse_next_response()

    def has_more_playlists(self):
        return self.continuation is not None

    async def async_create(self):
        self.prepare_request()
        self.data = await self.asyncPostRequest()
        self.parse_response()



from typing import List, Union

from py_yt.core.constants import (
    videoElementKey,
    channelElementKey,
    playlistElementKey,
    shelfElementKey,
)


class ComponentHandler:
    def _getVideoComponent(self, element: dict, shelfTitle: str = None) -> dict:
        video = element[videoElementKey]
        component = {
            "type": "video",
            "id": self._getValue(video, ["videoId"]),
            "title": self._getValue(video, ["title", "runs", 0, "text"]),
            "publishedTime": self._getValue(video, ["publishedTimeText", "simpleText"]),
            "duration": self._getValue(video, ["lengthText", "simpleText"]),
            "viewCount": {
                "text": self._getValue(video, ["viewCountText", "simpleText"]),
                "short": self._getValue(video, ["shortViewCountText", "simpleText"]),
            },
            "thumbnails": self._getValue(video, ["thumbnail", "thumbnails"]),
            "richThumbnail": self._getValue(
                video,
                [
                    "richThumbnail",
                    "movingThumbnailRenderer",
                    "movingThumbnailDetails",
                    "thumbnails",
                    0,
                ],
            ),
            "descriptionSnippet": self._getValue(
                video, ["detailedMetadataSnippets", 0, "snippetText", "runs"]
            ),
            "channel": {
                "name": self._getValue(video, ["ownerText", "runs", 0, "text"]),
                "id": self._getValue(
                    video,
                    [
                        "ownerText",
                        "runs",
                        0,
                        "navigationEndpoint",
                        "browseEndpoint",
                        "browseId",
                    ],
                ),
                "thumbnails": self._getValue(
                    video,
                    [
                        "channelThumbnailSupportedRenderers",
                        "channelThumbnailWithLinkRenderer",
                        "thumbnail",
                        "thumbnails",
                    ],
                ),
            },
            "accessibility": {
                "title": self._getValue(
                    video, ["title", "accessibility", "accessibilityData", "label"]
                ),
                "duration": self._getValue(
                    video, ["lengthText", "accessibility", "accessibilityData", "label"]
                ),
            },
        }
        component["link"] = "https://www.youtube.com/watch?v=" + component["id"]
        if component["channel"]["id"]:
            component["channel"]["link"] = (
                "https://www.youtube.com/channel/" + component["channel"]["id"]
            )
        component["shelfTitle"] = shelfTitle
        return component

    def _getChannelComponent(self, element: dict) -> dict:
        channel = element[channelElementKey]
        component = {
            "type": "channel",
            "id": self._getValue(channel, ["channelId"]),
            "title": self._getValue(channel, ["title", "simpleText"]),
            "thumbnails": self._getValue(channel, ["thumbnail", "thumbnails"]),
            "videoCount": self._getValue(
                channel, ["videoCountText", "runs", 0, "text"]
            ),
            "descriptionSnippet": self._getValue(
                channel, ["descriptionSnippet", "runs"]
            ),
            "subscribers": self._getValue(
                channel, ["subscriberCountText", "simpleText"]
            ),
        }
        component["link"] = "https://www.youtube.com/channel/" + component["id"]
        return component

    def _getPlaylistComponent(self, element: dict) -> dict:
        playlist = element[playlistElementKey]
        component = {
            "type": "playlist",
            "id": self._getValue(playlist, ["playlistId"]),
            "title": self._getValue(playlist, ["title", "simpleText"]),
            "videoCount": self._getValue(playlist, ["videoCount"]),
            "channel": {
                "name": self._getValue(
                    playlist, ["shortBylineText", "runs", 0, "text"]
                ),
                "id": self._getValue(
                    playlist,
                    [
                        "shortBylineText",
                        "runs",
                        0,
                        "navigationEndpoint",
                        "browseEndpoint",
                        "browseId",
                    ],
                ),
            },
            "thumbnails": self._getValue(
                playlist,
                [
                    "thumbnailRenderer",
                    "playlistVideoThumbnailRenderer",
                    "thumbnail",
                    "thumbnails",
                ],
            ),
        }
        component["link"] = "https://www.youtube.com/playlist?list=" + component["id"]
        if component["channel"]["id"]:
            component["channel"]["link"] = (
                "https://www.youtube.com/channel/" + component["channel"]["id"]
            )
        return component

    def _getVideoFromChannelSearch(self, elements: list) -> list:
        channelsearch = []
        for element in elements:
            element = self._getValue(element, ["childVideoRenderer"])
            json = {
                "id": self._getValue(element, ["videoId"]),
                "title": self._getValue(element, ["title", "simpleText"]),
                "uri": self._getValue(
                    element,
                    [
                        "navigationEndpoint",
                        "commandMetadata",
                        "webCommandMetadata",
                        "url",
                    ],
                ),
                "duration": {
                    "simpleText": self._getValue(element, ["lengthText", "simpleText"]),
                    "text": self._getValue(
                        element,
                        ["lengthText", "accessibility", "accessibilityData", "label"],
                    ),
                },
            }
            channelsearch.append(json)
        return channelsearch

    def _getChannelSearchComponent(self, elements: list) -> list:
        channelsearch = []
        for element in elements:
            responsetype = None

            if "gridPlaylistRenderer" in element:
                element = element["gridPlaylistRenderer"]
                responsetype = "gridplaylist"
            elif "itemSectionRenderer" in element:
                first_content = element["itemSectionRenderer"]["contents"][0]
                if "videoRenderer" in first_content:
                    element = first_content["videoRenderer"]
                    responsetype = "video"
                elif "playlistRenderer" in first_content:
                    element = first_content["playlistRenderer"]
                    responsetype = "playlist"
                else:
                    raise Exception(f"Unexpected first_content {first_content}")
            elif "continuationItemRenderer" in element:
                # for endless scrolling, not needed here
                # TODO: Implement endless scrolling
                continue
            else:
                raise Exception(f"Unexpected element {element}")

            if responsetype == "video":
                json = {
                    "id": self._getValue(element, ["videoId"]),
                    "thumbnails": {
                        "normal": self._getValue(element, ["thumbnail", "thumbnails"]),
                        "rich": self._getValue(
                            element,
                            [
                                "richThumbnail",
                                "movingThumbnailRenderer",
                                "movingThumbnailDetails",
                                "thumbnails",
                            ],
                        ),
                    },
                    "title": self._getValue(element, ["title", "runs", 0, "text"]),
                    "descriptionSnippet": self._getValue(
                        element, ["descriptionSnippet", "runs", 0, "text"]
                    ),
                    "uri": self._getValue(
                        element,
                        [
                            "navigationEndpoint",
                            "commandMetadata",
                            "webCommandMetadata",
                            "url",
                        ],
                    ),
                    "views": {
                        "precise": self._getValue(
                            element, ["viewCountText", "simpleText"]
                        ),
                        "simple": self._getValue(
                            element, ["shortViewCountText", "simpleText"]
                        ),
                        "approximate": self._getValue(
                            element,
                            [
                                "shortViewCountText",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                    },
                    "duration": {
                        "simpleText": self._getValue(
                            element, ["lengthText", "simpleText"]
                        ),
                        "text": self._getValue(
                            element,
                            [
                                "lengthText",
                                "accessibility",
                                "accessibilityData",
                                "label",
                            ],
                        ),
                    },
                    "published": self._getValue(
                        element, ["publishedTimeText", "simpleText"]
                    ),
                    "channel": {
                        "name": self._getValue(
                            element, ["ownerText", "runs", 0, "text"]
                        ),
                        "thumbnails": self._getValue(
                            element,
                            [
                                "channelThumbnailSupportedRenderers",
                                "channelThumbnailWithLinkRenderer",
                                "thumbnail",
                                "thumbnails",
                            ],
                        ),
                    },
                    "type": responsetype,
                }
            elif responsetype == "playlist":
                json = {
                    "id": self._getValue(element, ["playlistId"]),
                    "videos": self._getVideoFromChannelSearch(
                        self._getValue(element, ["videos"])
                    ),
                    "thumbnails": {
                        "normal": self._getValue(element, ["thumbnails"]),
                    },
                    "title": self._getValue(element, ["title", "simpleText"]),
                    "uri": self._getValue(
                        element,
                        [
                            "navigationEndpoint",
                            "commandMetadata",
                            "webCommandMetadata",
                            "url",
                        ],
                    ),
                    "channel": {
                        "name": self._getValue(
                            element, ["longBylineText", "runs", 0, "text"]
                        ),
                    },
                    "type": responsetype,
                }
            else:
                json = {
                    "id": self._getValue(element, ["playlistId"]),
                    "thumbnails": {
                        "normal": self._getValue(
                            element, ["thumbnail", "thumbnails", 0]
                        ),
                    },
                    "title": self._getValue(element, ["title", "runs", 0, "text"]),
                    "uri": self._getValue(
                        element,
                        [
                            "navigationEndpoint",
                            "commandMetadata",
                            "webCommandMetadata",
                            "url",
                        ],
                    ),
                    "type": "playlist",
                }
            channelsearch.append(json)
        return channelsearch

    def _getShelfComponent(self, element: dict) -> dict:
        shelf = element[shelfElementKey]
        return {
            "title": self._getValue(shelf, ["title", "simpleText"]),
            "elements": self._getValue(
                shelf, ["content", "verticalListRenderer", "items"]
            ),
        }

    def _getValue(self, source: dict, path: List[str]) -> Union[str, int, dict, None]:
        value = source
        for key in path:
            if type(key) is str:
                if key in value.keys():
                    value = value[key]
                else:
                    value = None
                    break
            elif type(key) is int:
                if len(value) != 0:
                    value = value[key]
                else:
                    value = None
                    break
        return value



import os

import httpx
from py_yt.core.constants import userAgent

from dotenv import load_dotenv

load_dotenv()


class RequestCore:
    def __init__(self, timeout: float = 3.0):
        self.url: str | None = None
        self.data: dict | None = None
        self.timeout: float = timeout
        self.proxy_url: str | None = os.environ.get("PROXY_URL")
        client_args = {"timeout": self.timeout, "proxy": self.proxy_url}

        self.async_client = httpx.AsyncClient(**client_args)

    async def asyncPostRequest(self) -> httpx.Response | None:
        """Sends an asynchronous POST request."""
        if not self.url:
            raise ValueError("URL must be set before making a request.")
        try:
            response = await self.async_client.post(
                self.url,
                headers={"User-Agent": userAgent},
                json=self.data,
            )
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            print(f"HTTP error: {e.response.status_code} - {e.response.text}")
        except httpx.RequestError as e:
            print(f"Request error: {e}")
        return None

    async def asyncGetRequest(self) -> httpx.Response | None:
        """Sends an asynchronous GET request."""
        if not self.url:
            raise ValueError("URL must be set before making a request.")
        cookies = {"CONSENT": "YES+1"}
        try:
            response = await self.async_client.get(
                self.url,
                headers={"User-Agent": userAgent},
                cookies=cookies,
            )
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            print(f"HTTP error: {e.response.status_code} - {e.response.text}")
        except httpx.RequestError as e:
            print(f"Request error: {e}")
        return None



import copy
import json
from typing import Union
from urllib.parse import urlencode

from py_yt.core.componenthandler import getVideoId, getValue
from py_yt.core.constants import searchKey, ResultMode
from py_yt.core.requests import RequestCore

CLIENTS = {
    "MWEB": {
        "context": {
            "client": {"clientName": "MWEB", "clientVersion": "2.20240425.01.00"}
        },
        "api_key": "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
    },
    "ANDROID": {
        "context": {"client": {"clientName": "ANDROID", "clientVersion": "19.02.39"}},
        "api_key": "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
    },
    "ANDROID_EMBED": {
        "context": {
            "client": {
                "clientName": "ANDROID",
                "clientVersion": "19.02.39",
                "clientScreen": "EMBED",
            }
        },
        "api_key": "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
    },
    "TV_EMBED": {
        "context": {
            "client": {
                "clientName": "TVHTML5_SIMPLY_EMBEDDED_PLAYER",
                "clientVersion": "2.0",
            },
            "thirdParty": {
                "embedUrl": "https://www.youtube.com/",
            },
        },
        "api_key": "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
    },
}


class VideoCore(RequestCore):
    def __init__(
        self,
        video_link: str,
        component_mode: str,
        result_mode: int,
        timeout: int,
        enable_html: bool,
        overrided_client: str = "ANDROID",
    ):
        super().__init__()
        self.timeout = timeout
        self.resultMode = result_mode
        self.componentMode = component_mode
        self.videoLink = video_link
        self.enableHTML = enable_html
        self.overridedClient = overrided_client

    # We call this when we use only HTML
    def post_request_only_html_processing(self):
        self.__getVideoComponent(self.componentMode)
        self.result = self.__videoComponent

    def post_request_processing(self):
        self.__parseSource()
        self.__getVideoComponent(self.componentMode)
        self.result = self.__videoComponent

    def prepare_innertube_request(self):
        self.url = (
            "https://www.youtube.com/youtubei/v1/player"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                    "contentCheckOk": True,
                    "racyCheckOk": True,
                    "videoId": getVideoId(self.videoLink),
                }
            )
        )
        self.data = copy.deepcopy(CLIENTS[self.overridedClient])

    async def async_create(self):
        self.prepare_innertube_request()
        response = await self.asyncPostRequest()
        if response is None:
            raise Exception("The request returned an empty response.")
        self.response = response.text
        if response.status_code == 200:
            self.post_request_processing()
        else:
            raise Exception("ERROR: Invalid status code.")

    def prepare_html_request(self):
        self.url = (
            "https://www.youtube.com/youtubei/v1/player"
            + "?"
            + urlencode(
                {
                    "key": searchKey,
                    "contentCheckOk": True,
                    "racyCheckOk": True,
                    "videoId": getVideoId(self.videoLink),
                }
            )
        )
        self.data = CLIENTS["MWEB"]

    async def async_html_create(self):
        self.prepare_html_request()
        response = await self.asyncPostRequest()
        self.HTMLresponseSource = response.json()

    def __parseSource(self) -> None:
        try:
            self.responseSource = json.loads(self.response)
        except Exception as e:
            raise Exception("ERROR: Could not parse YouTube response." + str(e))

    def __result(self, mode: int) -> Union[dict, str]:
        if mode == ResultMode.dict:
            return self.__videoComponent
        elif mode == ResultMode.json:
            return json.dumps(self.__videoComponent, indent=4)

    def __getVideoComponent(self, mode: str) -> None:
        videoComponent = {}
        if mode in ["getInfo", None]:
            responseSource = getattr(self, "responseSource", None)
            if self.enableHTML:
                responseSource = self.HTMLresponseSource
            component = {
                "id": getValue(responseSource, ["videoDetails", "videoId"]),
                "title": getValue(responseSource, ["videoDetails", "title"]),
                "duration": {
                    "secondsText": getValue(
                        responseSource, ["videoDetails", "lengthSeconds"]
                    ),
                },
                "viewCount": {
                    "text": getValue(responseSource, ["videoDetails", "viewCount"])
                },
                "thumbnails": getValue(
                    responseSource, ["videoDetails", "thumbnail", "thumbnails"]
                ),
                "description": getValue(
                    responseSource, ["videoDetails", "shortDescription"]
                ),
                "channel": {
                    "name": getValue(responseSource, ["videoDetails", "author"]),
                    "id": getValue(responseSource, ["videoDetails", "channelId"]),
                },
                "allowRatings": getValue(
                    responseSource, ["videoDetails", "allowRatings"]
                ),
                "averageRating": getValue(
                    responseSource, ["videoDetails", "averageRating"]
                ),
                "keywords": getValue(responseSource, ["videoDetails", "keywords"]),
                "isLiveContent": getValue(
                    responseSource, ["videoDetails", "isLiveContent"]
                ),
                "publishDate": getValue(
                    responseSource,
                    ["microformat", "playerMicroformatRenderer", "publishDate"],
                ),
                "uploadDate": getValue(
                    responseSource,
                    ["microformat", "playerMicroformatRenderer", "uploadDate"],
                ),
                "isFamilySafe": getValue(
                    responseSource,
                    ["microformat", "playerMicroformatRenderer", "isFamilySafe"],
                ),
                "category": getValue(
                    responseSource,
                    ["microformat", "playerMicroformatRenderer", "category"],
                ),
            }
            component["isLiveNow"] = (
                component["isLiveContent"]
                and component["duration"]["secondsText"] == "0"
            )
            component["link"] = "https://www.youtube.com/watch?v=" + component["id"]
            component["channel"]["link"] = (
                "https://www.youtube.com/channel/" + component["channel"]["id"]
            )
            videoComponent.update(component)
        if mode in ["getFormats", None]:
            videoComponent.update(
                {"streamingData": getValue(self.responseSource, ["streamingData"])}
            )
        if self.enableHTML:
            videoComponent["publishDate"] = getValue(
                self.HTMLresponseSource,
                ["microformat", "playerMicroformatRenderer", "publishDate"],
            )
            videoComponent["uploadDate"] = getValue(
                self.HTMLresponseSource,
                ["microformat", "playerMicroformatRenderer", "uploadDate"],
            )
        self.__videoComponent = videoComponent



import os

from setuptools import setup, find_packages

base_path = os.path.abspath(os.path.dirname(__file__))

about: dict = {}

with open(
    os.path.join(
        base_path,
        "py_yt",
        "__version__.py",
    ),
    encoding="utf-8",
) as f:
    exec(f.read(), about)

DESCRIPTION = (
    "A Python package for searching and retrieving YouTube data using py-yt-search."
)
with open("README.md", encoding="utf8") as readme:
    long_description = readme.read()

setup(
    name="py-yt-search",
    version=about["__version__"],
    author="AshokShau",
    author_email="<abishnoi69@outlook.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "httpx>=0.28.1",
        "python-dotenv>=1.1.0",
    ],
    keywords="youtube youtube-api video-search playlist channel search py_yt py-yt-search py-yt",
    url="https://github.com/AshokShau/py-yt",
    download_url="https://github.com/AshokShau/py-yt-search/releases/latest",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "License :: OSI Approved :: MIT License",
        "Topic :: Internet",
        "Topic :: Multimedia :: Video",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Source": "https://github.com/AshokShau/py-yt-search",
    },
    python_requires=">=3.8",
)
